@@ -32,6 +32,8 @@
 
 #include <locale.h>
 
+#include <dbus/dbus.h>
+
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <gdk/gdk.h>
@@ -45,6 +47,8 @@
 #include "gnome-settings-profile.h"
 #include "gsd-background-manager.h"
 
+#define NAUTILUS_SHOW_DESKTOP_KEY "/apps/nautilus/preferences/show_desktop"
+
 #define GSD_BACKGROUND_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_BACKGROUND_MANAGER, GsdBackgroundManagerPrivate))
 
 struct GsdBackgroundManagerPrivate
@@ -53,6 +57,8 @@ struct GsdBackgroundManagerPrivate
         GnomeBG     *bg;
         guint        bg_notify_id;
         guint        timeout_id;
+
+        DBusConnection *dbus_connection;
 };
 
 static void     gsd_background_manager_class_init  (GsdBackgroundManagerClass *klass);
@@ -155,7 +161,8 @@ nautilus_is_running (void)
 }
 
 static void
-draw_background (GsdBackgroundManager *manager)
+draw_background (GsdBackgroundManager *manager,
+                 gboolean              use_crossfade)
 {
         GdkDisplay *display;
         int         n_screens;
@@ -185,7 +192,15 @@ draw_background (GsdBackgroundManager *manager)
                                                  gdk_screen_get_height (screen),
                                                  TRUE);
 
-                gnome_bg_set_pixmap_as_root (screen, pixmap);
+                if (use_crossfade) {
+                        GnomeBGCrossfade *fade;
+
+                        fade = gnome_bg_set_pixmap_as_root_with_crossfade (screen, pixmap);
+                        g_signal_connect (fade, "finished",
+                                          G_CALLBACK (g_object_unref), NULL);
+                } else {
+                        gnome_bg_set_pixmap_as_root (screen, pixmap);
+                }
 
                 g_object_unref (pixmap);
         }
@@ -193,19 +208,18 @@ draw_background (GsdBackgroundManager *manager)
         gnome_settings_profile_end (NULL);
 }
 
-static gboolean
-queue_draw_background (GsdBackgroundManager *manager)
-{
-        manager->priv->timeout_id = 0;
-        draw_background (manager);
-        return FALSE;
-}
-
 static void
 on_bg_changed (GnomeBG              *bg,
                GsdBackgroundManager *manager)
 {
-        draw_background (manager);
+        draw_background (manager, TRUE);
+}
+
+static void
+on_bg_transitioned (GnomeBG              *bg,
+                    GsdBackgroundManager *manager)
+{
+        draw_background (manager, FALSE);
 }
 
 static void
@@ -235,6 +249,150 @@ watch_bg_preferences (GsdBackgroundManager *manager)
                                                                NULL);
 }
 
+static void
+setup_bg (GsdBackgroundManager *manager)
+{
+        g_return_if_fail (manager->priv->bg == NULL);
+
+        manager->priv->bg = gnome_bg_new ();
+
+        g_signal_connect (manager->priv->bg,
+                          "changed",
+                          G_CALLBACK (on_bg_changed),
+                          manager);
+
+        g_signal_connect (manager->priv->bg,
+                          "transitioned",
+                          G_CALLBACK (on_bg_transitioned),
+                          manager);
+
+        watch_bg_preferences (manager);
+        gnome_bg_load_from_preferences (manager->priv->bg,
+                                        manager->priv->client);
+}
+
+static gboolean
+queue_draw_background (GsdBackgroundManager *manager)
+{
+        manager->priv->timeout_id = 0;
+        if (nautilus_is_running ()) {
+                return FALSE;
+        }
+        setup_bg (manager);
+        draw_background (manager, FALSE);
+        return FALSE;
+}
+
+static DBusHandlerResult
+on_bus_message (DBusConnection *connection,
+                DBusMessage    *message,
+                void           *user_data)
+{
+        GsdBackgroundManager *manager = user_data;
+
+        if (dbus_message_is_signal (message,
+                                    "org.gnome.SessionManager",
+                                    "SessionRunning")) {
+                /* If the session finishes then check if nautilus is
+                 * running and if not, set the background.
+                 *
+                 * We wait a few seconds after the session is up
+                 * because nautilus tells the session manager that its
+                 * ready before it sets the background.
+                 */
+                manager->priv->timeout_id = g_timeout_add_seconds (8,
+                                                                   (GSourceFunc)
+                                                                   queue_draw_background,
+                                                                   manager);
+                dbus_connection_remove_filter (connection,
+                                               on_bus_message,
+                                               manager);
+
+                manager->priv->dbus_connection = NULL;
+        }
+
+        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
+static void
+draw_background_after_session_loads (GsdBackgroundManager *manager)
+{
+        DBusConnection *connection;
+
+        connection = dbus_bus_get (DBUS_BUS_SESSION, NULL);
+
+        if (connection == NULL) {
+                return;
+        }
+
+        if (!dbus_connection_add_filter (connection, on_bus_message, manager, NULL)) {
+                return;
+        };
+
+        manager->priv->dbus_connection = connection;
+}
+
+static void
+on_screen_size_changed (GdkScreen            *screen,
+                        GsdBackgroundManager *manager)
+{
+        gboolean nautilus_show_desktop;
+
+        nautilus_show_desktop = gconf_client_get_bool (manager->priv->client,
+                                                       NAUTILUS_SHOW_DESKTOP_KEY,
+                                                       NULL);
+
+        if (!nautilus_is_running () || !nautilus_show_desktop) {
+                if (manager->priv->bg == NULL) {
+                        setup_bg (manager);
+                }
+                draw_background (manager, FALSE);
+        }
+}
+
+static void
+disconnect_screen_signals (GsdBackgroundManager *manager)
+{
+        GdkDisplay *display;
+        int         i;
+        int         n_screens;
+
+        display = gdk_display_get_default ();
+        n_screens = gdk_display_get_n_screens (display);
+
+        for (i = 0; i < n_screens; ++i) {
+                GdkScreen *screen;
+                screen = gdk_display_get_screen (display, i);
+                g_signal_handlers_disconnect_by_func (screen,
+                                                      G_CALLBACK (on_screen_size_changed),
+                                                      manager);
+        }
+}
+
+static void
+connect_screen_signals (GsdBackgroundManager *manager)
+{
+        GdkDisplay *display;
+        int         i;
+        int         n_screens;
+
+        display = gdk_display_get_default ();
+        n_screens = gdk_display_get_n_screens (display);
+
+        for (i = 0; i < n_screens; ++i) {
+                GdkScreen *screen;
+                screen = gdk_display_get_screen (display, i);
+                g_signal_connect (screen,
+                                  "monitors-changed",
+                                  G_CALLBACK (on_screen_size_changed),
+                                  manager);
+                g_signal_connect (screen,
+                                  "size-changed",
+                                  G_CALLBACK (on_screen_size_changed),
+                                  manager);
+        }
+}
+
 gboolean
 gsd_background_manager_start (GsdBackgroundManager *manager,
                               GError              **error)
@@ -245,16 +403,6 @@ gsd_background_manager_start (GsdBackgroundManager *manager,
         gnome_settings_profile_start (NULL);
 
         manager->priv->client = gconf_client_get_default ();
-        manager->priv->bg = gnome_bg_new ();
-
-        g_signal_connect (manager->priv->bg,
-                          "changed",
-                          G_CALLBACK (on_bg_changed),
-                          manager);
-
-        watch_bg_preferences (manager);
-        gnome_bg_load_from_preferences (manager->priv->bg,
-                                        manager->priv->client);
 
         /* If this is set, nautilus will draw the background and is
 	 * almost definitely in our session.  however, it may not be
@@ -264,18 +412,17 @@ gsd_background_manager_start (GsdBackgroundManager *manager,
 	 * nautilus overwrite it.
 	 */
         nautilus_show_desktop = gconf_client_get_bool (manager->priv->client,
-                                                       "/apps/nautilus/preferences/show_desktop",
+                                                       NAUTILUS_SHOW_DESKTOP_KEY,
                                                        NULL);
 
         if (!nautilus_show_desktop) {
-                draw_background (manager);
+                setup_bg (manager);
         } else {
-                /* even when nautilus is supposedly handling the
-                 * background, apply the settings eventually to make
-                 * people running a nautilus-less session happy */
-                manager->priv->timeout_id = g_timeout_add_seconds (8, (GSourceFunc)queue_draw_background, manager);
+                draw_background_after_session_loads (manager);
         }
 
+        connect_screen_signals (manager);
+
         gnome_settings_profile_end (NULL);
 
         return TRUE;
@@ -288,6 +435,14 @@ gsd_background_manager_stop (GsdBackgroundManager *manager)
 
         g_debug ("Stopping background manager");
 
+        disconnect_screen_signals (manager);
+
+        if (manager->priv->dbus_connection != NULL) {
+                dbus_connection_remove_filter (manager->priv->dbus_connection,
+                                               on_bus_message,
+                                               manager);
+        }
+
         if (manager->priv->bg_notify_id != 0) {
                 gconf_client_remove_dir (manager->priv->client,
                                          GNOME_BG_KEY_DIR,
